#include <avr/io.h>
#include <avr/interrupt.h> 
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <util/delay.h>
#include <NewSoftSerial.h>
#include "a034.h"


#define USART_BAUDRATE 8192
#define BAUD_PRESCALE (((F_CPU / (USART_BAUDRATE * 16UL))) - 1) 

NewSoftSerial mySerial(2, 3);
char dmode=0;        //input to change display mode
bool inSync=false;  //Track if data is in sync
char version[13] = "201110011659";
#define   RXcontrol   5        //Pin # RX enable transistor

int mdelay=3380;

#define Bit_Set(port,bit_num) (port = port | (0x01 << bit_num))
#define Bit_Clear(port,bit_num) (port = port & ~(0x01 << bit_num))

#define CarDataSize     22    //Size of ASCII car data

ALDL aldl(13);

//ECU trouble codes lookup table
const unsigned char PROGMEM ErrorCode_Table[]=
{
	24,23,22,21,15,14,13,12,
	42,41,35,34,33,32,31,25,
	55,54,53,26,51,45,44,43
};

//ISR(USART0_RXC_vect)
//{
//Interrupt tutorial from:
//http://www.avrfreaks.net/index.php?name=PNphpBB2&file=viewtopic&t=48188
//	DataStreamCounter++;
//DataStream[DataStreamCounter] = UDR0;
// Code to be executed when the USART receives a byte here
//} 

void setup()  
{
	cli();  //diable interrupts until ready
	//pinMode(5, OUTPUT);
	//set PD5 high to transmit aldl
	//digitalWrite(5,LOW);
	//DDRC = 0xFF;                                 // Configure PORTC as output
	DDRD = (1<<1)|(1<<3)|(1<<5)|(1<<6)|(1<<7);  // Configure PORTD 

	USART_init();
	//Serial.begin(8192);
	//Serial.println("Goodnight moon!");

	// set the data rate for the NewSoftSerial port
	mySerial.begin(57600);
	//mySerial.print(FreeRam());
	mySerial.println();
	mySerial.print("ALDL Info\tversion:");
	mySerial.print(version);
	mySerial.print("\tCompiled on ");
	mySerial.print(__DATE__);
	mySerial.print(" at ");
	mySerial.print(__TIME__);
	mySerial.println();
	sei();  //enable interrupts
}

void loop()
{
	if (mySerial.available()) {
		dmode = mySerial.read();
	}
	if (inSync == false) {
		mySerial.println(mdelay);
	}
	//if (DataStreamCount > 0) {
	//	process_car();
	//} else if (DataStreamCount > DataStreamSize - 2) {
	//	DataStreamCount = 0;
	//}
	Get_Car_Raw();
	aldl.calc_data();
	out_data2();
	if (dmode == '1') {
		Output_DataStream();
		mySerial.println(millis());
	} else if (dmode == '2') {
		Output_DataStreamc();
		mySerial.println(millis());
	} else if (dmode == 'c') {
		Output_CarDatac();
		mySerial.println(millis());
	}

}

//******************************************************************
//* Function: USART_init                                           *
//* Enable and Initialize UART                                     *
//******************************************************************
void USART_init(void)
{
	// Set the USART baudrate registers for 8192, UBRR=149
	UBRR0H = (BAUD_PRESCALE >> 8); // Load upper 8-bits of the baud rate value into the high byte of the UBRR register 
	UBRR0L = BAUD_PRESCALE; // Load lower 8-bits of the baud rate value into the low byte of the UBRR register

	//UCSRnA - USART Control and Status Register A
	// Enable 2x speed change, 0=16 divider 1=8 divider
	/*
	Bit 7 MSB
	R RXCn  -> USART Receive Complete
	0 TXCn  -> TXCn: USART Transmit Complete
	R UDREn -> USART Data Register Empty
	R FEn   -> Frame Error
	R DORn  -> Data OverRun
	R UPEn  -> USART Parity Error
	0 U2Xn  -> Double the USART Transmission Speed, 0=16 divider 1=8 divider
	0 MPCMn -> Multi-processor Communication Mode
	Bit 0 LSB
	*/
	UCSR0A = (0<<U2X0);


	// UCSRnB - USART Control and Status Register n B
	// Enable receiver
	/*
	Bit 7 MSB
	RXCIE0 = 1 Receive Interrupt enabled
	TXCIE0 = 0 Transmit Interrupt disabled
	UDRIE0 = 0 Interupt disabled
	RXEN0  = 1 Receive enabled
	TXEN0  = 1 Transmit enabled
	UCSZ02 = 0 8-bit mode
	RXB80  = 0 9'th RX data bit when using 9-bit UART
	TXB80  = 0 9'th TX data bit when using 9-bit UART
	Bit 0 LSB
	*/
	UCSR0B =(1<<RXEN0)|(1<<TXEN0);  //|(1<<RXCIE0); When we move to interrupt based receive

	
	// UCSRnC - USART Control and Status Register n C
	// Set the USART to asynchronous at 8 bits no parity and 1 stop bits
	/*
	Bit 7 MSB
	0 UMSEL01 |
	0 UMSEL00 |-> Async
	0 UPM01  |
	0 UPM00  |-> No parity
	0 USBS0 -> 1 stop bits;
	1 UCSZ01  |
	1 UCSZ00  |-> 8-bit
	0 UCPOL0 -> Reciving data sampled on fallign edge
	Bit 0 LSB
	*/
	UCSR0C =(0<<UMSEL01)|(0<<UMSEL00)|(0<<UPM01)|(0<<UPM00)|(0<<USBS0)|(1<<UCSZ01)|(1<<UCSZ00)|(0<<UCPOL0);
	
}


//*******************************************************************
//* Function: USART_Receive                                         *
//* Waits for incoming data on USART then returns it to the calling *
//* function                                                        *
//* - Commented out, not used in final working code -               *
//*******************************************************************

unsigned char USART_Receive(void)
{
	// Wait for data to be received 
	while ( !(UCSR0A & (1<<RXC0)) );

	// Get and return received data from buffer
	return UDR0;
}


//******************************************************************
//* Function: USART_Transmit                                       *
//* Transmit data                                                  *
//* datastream array until array is full                           *
//******************************************************************
void USART_Transmit(unsigned char data)
{
	// Wait for empty transmit buffer
	while ( !( UCSR0A & (1<<UDRE0)) );

	// Put data into buffer, sends the data
	UDR0 = data;

	// Wait for empty transmit buffer
	while ( !( UCSR0A & (1<<UDRE0)) );
}


//******************************************************************
//* Function: Get_Car_Data                                         *
//* Waits for incoming data on USART then puts received char into  *
//* datastream array until array is full                           *
//******************************************************************
int Get_Car_Raw(void)
{
	unsigned int timeout;

	if (inSync == false) {
		Bit_Clear(PORTD,RXcontrol);  //Open up recieve pin
		while (USART_Receive() != 0xF0) {}
		//while ( !(UCSR0A & (1<<RXC0)) );
		Bit_Set(PORTD,RXcontrol);  //Lockout recieve pin
		delayMicroseconds(mdelay);
	}

	Bit_Set(PORTD,RXcontrol);  //Lockout recieve pin

	//Mode 1 command set

	USART_Transmit(0xF0);
	USART_Transmit(0x57);
	USART_Transmit(0x01);
	USART_Transmit(0x00);
	USART_Transmit(0xB8);
	/*
	Serial.print(0xF4);
	Serial.print(0x56);
	Serial.print(0x01);
	Serial.print(0xB5);
	*/
	//delayMicroseconds(976);

	Bit_Clear(PORTD,RXcontrol);  //Open up recieve pin

	for(unsigned char i = 0 ; i <= DataStreamSize ; i++)
	{
		timeout=0;  // Reset timeout to zero

		// Wait for data to be received if data not recived in timout period
		// then exit loop
		while( !(UCSR0A & (1<<RXC0) )  && (timeout < 9000) )
		{
			timeout++;
		}

		// Get and return received data from buffer
		if(timeout == 9000)
		aldl.DataStream[i]=0;     // If timeout load 0 into datastream value
		else
		aldl.DataStream[i]=UDR0;  // Otherwise load recieved value
	}

	Bit_Set(PORTD,RXcontrol);  //Lockout recieve pin
	if ((aldl.DataStream[1] == 0xF0) && (aldl.DataStream[2] == 0x95)) {
		inSync=true;
		return 1;
	} else {
		if (inSync = false) {
			mdelay = mdelay + 1;
		}
		inSync=false;
		//		mySerial.print(DataStream[1],HEX);
		//		mySerial.print(DataStream[2],HEX);
		return 0;
	}
}



//******************************************************************
//* Function: Output_CarData()                                     *
//* Output calcualted car data and trouble codes to LCD            *
//******************************************************************
void Output_CarData()
{
	unsigned char ignore;
	unsigned int  ErrorCode;
	unsigned char bitn=0;
	unsigned char code_2, code_1;
	unsigned char error_bit, error_byte;


	error_bit=0;  //Reset error bit to zero


	//*************************
	// Check for Trouble Codes
	//*************************
	/*	for(error_byte = DEF_ERROR1 ; error_byte < DEF_ERROR3 ; error_byte++)
	{
	for(bitn = 0 ; bitn < 7 ; bitn++)
	{
		//If error bit in datastream is 1 then ouput trouble code
		if(DataStream[error_byte] & ( 1 << (bitn) ) )
		{
			ErrorCode = (unsigned char)pgm_read_byte(&ErrorCode_Table[error_bit]);
	
			if(ErrorCode != 0) //If error code is 0 ignore trouble code
			{
			ErrorCode = (unsigned char)pgm_read_byte(&ErrorCode_Table[error_bit]);
			
			HEXtoASCII(ErrorCode,&ignore, &ignore, &code_2, &code_1);
				mySerial.print(code_2);
			mySerial.print(code_1);
			mySerial.print(' ');
			}
		}

		error_bit = error_bit + 1; //Increment error bit number

	}
	} */
	mySerial.println();

}

void Output_DataStream(void) {
	for (int j=0; j<=DataStreamSize; j++) {
		mySerial.print(j);
		mySerial.print(":");
		mySerial.print(aldl.DataStream[j],HEX);
		mySerial.print(" ");
	}
}
void Output_DataStreamc(void) {
	for (int j=0; j<=DataStreamSize; j++) {
		//mySerial.print(j);
		//mySerial.print(":");
		mySerial.print(aldl.DataStream[j],HEX);
		mySerial.print(",");
	}
}
void Output_CarDatac(void) {
	for (int j=0; j<=CarDataInt; j++) {
		//mySerial.print(j);
		//mySerial.print(":");
		mySerial.print(aldl.CarData[j]);
		mySerial.print(",");
	}
}

void out_data(void) {
	aldl.disp_counter++;
	mySerial.print("RPM: ");
	mySerial.print(aldl.CarData[CAR_RPM]);
	mySerial.print(" MPH: ");
	mySerial.print(aldl.CarData[CAR_MPH]);
	mySerial.print(" TPS: ");
	mySerial.print(aldl.CarData[CAR_TPS]);
	mySerial.print(" Coolant: ");
	mySerial.print(aldl.CarData[CAR_CTS]);
	mySerial.print("C O2: ");
	mySerial.print(aldl.CarData[CAR_O2S]);
	mySerial.print(" SparkAdvance: ");
	mySerial.print(aldl.CarData[CAR_SPAD]);
	mySerial.print(" MAP,MAPLV: ");
	mySerial.print(aldl.CarData[CAR_MAP]);
	mySerial.print(",");
	mySerial.print(aldl.CarData[CAR_MAPLV]);
	mySerial.print(" RTPS: ");
	mySerial.print(aldl.CarData[CAR_RTPS]);
	mySerial.print(" Batt: ");
	mySerial.print(aldl.CarData[CAR_BATT]);
	mySerial.println("V");

}

void out_data2(void) {
	if (aldl.disp_counter > 25 || aldl.disp_counter == 0) {
		mySerial.println("RPM\tMPH\tTPS\tO2\tCool\tSpkA\tIACD\tMAP\tMAPLV\tVATT");
		aldl.disp_counter = 0;
	}
	mySerial.print(aldl.CarData[CAR_RPM]);
	mySerial.print("\t");
	mySerial.print(aldl.CarData[CAR_MPH]);
	mySerial.print("\t");
	mySerial.print(aldl.CarData[CAR_RTPS]);
	mySerial.print("\t");
	mySerial.print(aldl.CarData[CAR_O2S]);
	mySerial.print("\t");
	mySerial.print(aldl.CarData[CAR_CTS]);
	mySerial.print("C\t");
	mySerial.print(aldl.CarData[CAR_SPAD]);
	mySerial.print("\t");
	mySerial.print(aldl.CarData[CAR_IACDF]);
	mySerial.print("\t");
	mySerial.print(aldl.CarData[CAR_MAP]);
	mySerial.print("\t");
	mySerial.print(aldl.CarData[CAR_MAPLV]);
	mySerial.print("\t");
	mySerial.print(aldl.CarData[CAR_BATT]);
	mySerial.println("V");
	aldl.disp_counter++;

}

